"""Code for RandOm Convolutional KErnel Transformation."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

from numba import njit, prange
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils.validation import (
    check_array, check_is_fitted, check_random_state)
from ..base import UnivariateTransformerMixin


@njit()
def generate_kernels(n_kernels, n_timestamps, kernel_sizes, seed):
    """Generate the kernels.

    Parameters
    ----------
    n_kernels : int
        Number of kernels

    n_timestamps : int
        Number of timestamps

    kernel_sizes : array
        Possible sizes for the kernels.

    seed : int
        Seed for the random number generator.

    Returns
    -------
    weights : array, shape = (n_kernels, max(kernel_sizes))
        Weights of the kernels. Zero padding values are added.

    lengths : array, shape = (n_kernels,)
        Length of each kernel.

    biases : array, shape = (n_kernels,)
        Bias of each kernel.

    dilations : array, shape = (n_kernels,)
        Dilation of each kernel.

    paddings : array, shape = (n_kernels,)
        Padding of each kernel.

    """
    # Fix the random see
    np.random.seed(seed)

    # Lengths of the kernels
    lengths = np.random.choice(kernel_sizes, size=n_kernels)

    # Weights of the kernels
    cumsum_lengths = np.concatenate((np.array([0]), np.cumsum(lengths)))
    weights_all = np.random.randn(cumsum_lengths[-1])
    weights = np.zeros((n_kernels, np.int64(np.max(kernel_sizes))))
    for i in prange(n_kernels):
        weights[i, :lengths[i]] = (
            weights_all[cumsum_lengths[i]: cumsum_lengths[i+1]] -
            np.mean(weights_all[cumsum_lengths[i]: cumsum_lengths[i+1]])
        )

    # Biases
    biases = np.random.uniform(-1, 1, size=n_kernels)

    # Dilations
    upper_bounds = np.log2(np.floor_divide(n_timestamps - 1, lengths - 1))
    powers = np.empty(n_kernels)
    for i in prange(n_kernels):
        powers[i] = np.random.uniform(0, upper_bounds[i])
    dilations = np.floor(np.power(2, powers))

    # Paddings
    paddings = np.zeros(n_kernels)
    padding_cond = np.random.randint(0, 2, n_kernels).astype(np.bool_)
    paddings[padding_cond] = np.floor_divide(
        (lengths - 1) * dilations, 2)[padding_cond]

    return weights, lengths, biases, dilations, paddings


@njit(fastmath=True)
def apply_one_kernel_one_sample(
    x, n_timestamps, weight, length, bias, dilation, padding
):
    """Apply one kernel to one time series.

    Parameters
    ----------
    x : array, shape = (n_timestamps,)
        One time series.

    n_timestamps : int
        Number of timestamps.

    weights : array, shape = (length,)
        Weights of the kernel. Zero padding values are added.

    length : int
        Length of the kernel.

    bias : int
        Bias of the kernel.

    dilation : int
        Dilation of the kernel.

    padding : int
        Padding of the kernel.

    Returns
    -------
    x_new : array, shape = (2,)
        Extracted features using the kernel.

    """
    # Compute padded x
    n_conv = n_timestamps - ((length - 1) * dilation) + (2 * padding)
    if padding > 0:
        x_pad = np.zeros(n_timestamps + 2 * padding)
        x_pad[padding:-padding] = x
    else:
        x_pad = x

    # Compute the convolutions
    x_conv = np.zeros(n_conv)
    for i in prange(n_conv):
        for j in prange(length):
            x_conv[i] += weight[j] * x_pad[i + (j * dilation)]
    x_conv += bias

    # Return the features: maximum and proportion of positive values
    return np.max(x_conv), np.mean(x_conv > 0)


@njit()
def apply_all_kernels(X, weights, lengths, biases, dilations, paddings):
    """Apply one kernel to a data set of time series.

    Parameters
    ----------
    X : array, shape = (n_samples, n_timestamps)
        Input data.

    weights : array, shape = (n_kernels, max(kernel_sizes))
        Weights of the kernels. Zero padding values are added.

    lengths : array, shape = (n_kernels,)
        Length of each kernel.

    biases : array, shape = (n_kernels,)
        Bias of each kernel.

    dilations : array, shape = (n_kernels,)
        Dilation of each kernel.

    paddings : array, shape = (n_kernels,)
        Padding of each kernel.

    Returns
    -------
    X_new : array, shape = (n_samples, 2 * n_kernels)
        Extracted features using all the kernels.

    """
    n_samples, n_timestamps = X.shape
    n_kernels = lengths.size
    X_new = np.empty((n_samples, 2 * n_kernels))
    for i in prange(n_samples):
        for j in prange(n_kernels):
            X_new[i, (2 * j):(2 * j + 2)] = apply_one_kernel_one_sample(
                X[i], n_timestamps, weights[j], lengths[j],
                biases[j], dilations[j], paddings[j]
            )
    return X_new


class ROCKET(BaseEstimator, UnivariateTransformerMixin):
    """RandOm Convolutional KErnel Transformation.

    This algorithm randomly generates a great variety of convolutional kernels
    and extracts two features for each convolution: the maximum and the
    proportion of positive values.

    Parameters
    ----------
    n_kernels : int (default = 10000)
        Number of kernels.

    kernel_sizes : array-like (default = (7, 9, 11))
        The possible sizes of the kernels.

    random_state : None, int or RandomState instance (default = None)
        The seed of the pseudo random number generator to use when shuffling
        the data. If int, random_state is the seed used by the random number
        generator. If RandomState instance, random_state is the random number
        generator. If None, the random number generator is the RandomState
        instance used by `np.random`.

    Attributes
    ----------
    weights_ : array, shape = (n_kernels, max(kernel_sizes))
        Weights of the kernels. Zero padding values are added.

    length_ : array, shape = (n_kernels,)
        Length of each kernel.

    bias_ : array, shape = (n_kernels,)
        Bias of each kernel.

    dilation_ : array, shape = (n_kernels,)
        Dilation of each kernel.

    padding_ : array, shape = (n_kernels,)
        Padding of each kernel.

    References
    ----------
    .. [1] A. Dempster, F. Petitjean and G. I. Webb, "ROCKET: Exceptionally
           fast and accurate time series classification using random
           convolutional kernels". https://arxiv.org/abs/1910.13051.

    Examples
    --------
    >>> from pyts.transformation import ROCKET
    >>> X = np.arange(100).reshape(5, 20)
    >>> rocket = ROCKET(n_kernels=10)
    >>> rocket.fit_transform(X).shape
    (5, 20)

    """
    def __init__(self, n_kernels=10000, kernel_sizes=(7, 9, 11),
                 random_state=None):
        self.n_kernels = n_kernels
        self.kernel_sizes = kernel_sizes
        self.random_state = random_state

    def fit(self, X, y=None):
        """Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Training vector.

        y : None or array-like, shape = (n_samples,)
            Class labels for each data sample. Ignored.

        Returns
        -------
        self : object

        """
        X = check_array(X, dtype='float64')
        n_samples, n_timestamps = X.shape

        kernel_sizes, seed = self._check_params(n_timestamps)

        # Generate the kernels
        weights, lengths, biases, dilations, paddings = generate_kernels(
            self.n_kernels, n_timestamps, kernel_sizes, seed)

        self.weights_ = weights
        self.length_ = lengths
        self.bias_ = biases
        self.dilation_ = dilations.astype('int64')
        self.padding_ = paddings.astype('int64')

        return self

    def transform(self, X):
        """Transform the provided data.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)
            Test samples.

        Returns
        -------
        X_new : array, shape = (n_samples, 2 * n_kernels)
            Extracted features from the kernels.

        """
        X = check_array(X, dtype='float64')
        check_is_fitted(self, ['weights_', 'length_', 'bias_',
                               'dilation_', 'padding_'])
        X_new = apply_all_kernels(
            X, self.weights_, self.length_, self.bias_,
            self.dilation_, self.padding_
        )
        return X_new

    def _check_params(self, n_timestamps):
        if not isinstance(self.n_kernels, (int, np.integer)):
            raise TypeError("'n_kernels' must be an integer (got {})."
                            .format(self.n_kernels))

        if not isinstance(self.kernel_sizes, (list, tuple, np.ndarray)):
            raise TypeError("'kernel_sizes' must be a list, a tuple or "
                            "an array (got {}).".format(self.kernel_sizes))
        kernel_sizes = check_array(self.kernel_sizes, ensure_2d=False,
                                   dtype='int64', accept_large_sparse=False)
        if not np.all(1 <= kernel_sizes):
            raise ValueError("All the values in 'kernel_sizes' must be "
                             "greater than or equal to 1 ({} < 1)."
                             .format(kernel_sizes.min()))
        if not np.all(kernel_sizes <= n_timestamps):
            raise ValueError("All the values in 'kernel_sizes' must be lower "
                             "than or equal to 'n_timestamps' ({} > {})."
                             .format(kernel_sizes.max(), n_timestamps))

        rng = check_random_state(self.random_state)
        seed = rng.randint(np.iinfo(np.uint32).max, dtype='u8')

        return kernel_sizes, seed



"""Testing for 'segmentation' function."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import pytest
import re
from pyts.utils import segmentation


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'ts_size': None, 'window_size': 3}, TypeError,
      "'ts_size' must be an integer."),
     ({'ts_size': 4, 'window_size': None}, TypeError,
      "'window_size' must be an integer."),
     ({'ts_size': 4, 'window_size': 3, 'n_segments': "3"}, TypeError,
      "'n_segments' must be None or an integer."),
     ({'ts_size': 1, 'window_size': 2}, ValueError,
      "'ts_size' must be an integer greater than or equal to 2 (got 1)."),
     ({'ts_size': 10, 'window_size': 0}, ValueError,
      "'window_size' must be an integer greater than or equal to 1 (got 0)."),
     ({'ts_size': 10, 'window_size': 15}, ValueError,
      "'window_size' must be lower than or equal to 'ts_size' (15 > 10)."),
     ({'ts_size': 10, 'window_size': 3, 'n_segments': 1}, ValueError,
      "If 'n_segments' is an integer, it must be greater than or equal to 2 "
      "(got 1)."),
     ({'ts_size': 10, 'window_size': 3, 'n_segments': 12}, ValueError,
      "If 'n_segments' is an integer, it must be lower than or equal to "
      "'ts_size' (12 > 10).")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    with pytest.raises(error, match=re.escape(err_msg)):
        segmentation(**params)


@pytest.mark.parametrize(
    ('params, start_desired, end_desired, size_desired'),
    [({'ts_size': 20, 'window_size': 4, 'overlapping': False},
      [0, 4, 8, 12, 16], [4, 8, 12, 16, 20], 5),
     ({'ts_size': 20, 'window_size': 8, 'overlapping': True},
      [0, 6, 12], [8, 14, 20], 3)]
)
def test_accurate_results(params, start_desired, end_desired, size_desired):
    """Test that the actual results are the expected ones."""
    start_actual, end_actual, size_actual = segmentation(**params)
    np.testing.assert_array_equal(start_actual, start_desired)
    np.testing.assert_array_equal(end_actual, end_desired)
    np.testing.assert_equal(size_actual, size_desired)



import numpy as np
import pytest
import re

from pyts.transformation.rocket import (
    generate_kernels, apply_one_kernel_one_sample, apply_all_kernels)
from pyts.transformation import ROCKET


X = np.arange(14, dtype='float64').reshape(2, 7)
weights = np.array([[1, -1, 0, 0],
                    [-1, 2, -1, 0],
                    [-0.5, 0.5, 0, 0]])
lengths = np.array([2, 3, 4])
biases = np.array([-1., 0., 1.])
dilations = np.array([1, 2, 2])
paddings = np.array([0, 0, 3])


@pytest.mark.parametrize(
    'n_kernels, n_timestamps, kernel_sizes, seed',
    [(10, 100, (4, 6, 8), 42),
     (10, 100, (4, 6, 8), 43),
     (50, 100, (4, 6, 8), 42),
     (10, 64, (4, 6, 8), 42),
     (10, 600, (4, 6, 8), 42)]
)
def test_generate_kernels(n_kernels, n_timestamps, kernel_sizes, seed):
    """Test the generated kernels."""
    kernel_sizes = np.asarray(kernel_sizes)
    weights, lengths, biases, dilations, paddings = generate_kernels(
        n_kernels, n_timestamps, kernel_sizes, seed
    )
    # Check shapes
    assert weights.shape == (n_kernels, max(kernel_sizes))
    for arr in (lengths, biases, dilations, paddings):
        assert arr.shape == (n_kernels,)

    # Check range of values
    assert np.all(np.isin(lengths, kernel_sizes))
    assert np.all(np.logical_and(-1 <= biases, biases <= 1))
    (n_timestamps - 1) // (min(kernel_sizes) - 1)
    upper_bound_dilation = (n_timestamps - 1) // (min(kernel_sizes) - 1)
    assert np.all(
        np.logical_and(1 <= dilations, dilations <= upper_bound_dilation))
    upper_bound_padding = (
        (lengths - 1) * (n_timestamps - 1) // (min(kernel_sizes) - 1)) // 2
    assert np.all(
        np.logical_and(0 <= paddings, paddings <= upper_bound_padding))

    # Check zero mean of weights
    np.testing.assert_allclose(weights.mean(axis=1), 0, rtol=0, atol=1e-5)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'weight': np.zeros(2), 'length': 2, 'bias': 0,
       'dilation': 1, 'padding': 0}, [0, 0]),
     ({'weight': np.zeros(2), 'length': 2, 'bias': -1,
       'dilation': 1, 'padding': 0}, [-1, 0]),
     ({'weight': np.zeros(2), 'length': 2, 'bias': 1,
       'dilation': 1, 'padding': 0}, [1, 1]),
     ({'weight': np.zeros(2), 'length': 2, 'bias': 1,
       'dilation': 2, 'padding': 0}, [1, 1]),
     ({'weight': np.zeros(3), 'length': 2, 'bias': 1,
       'dilation': 1, 'padding': 0}, [1, 1]),
     ({'weight': np.zeros(3), 'length': 2, 'bias': 1,
       'dilation': 1, 'padding': 1}, [1, 1]),
     ({'weight': np.zeros(3), 'length': 2, 'bias': 1,
       'dilation': 2, 'padding': 2}, [1, 1]),
     ({'weight': weights[0], 'length': lengths[0],
       'bias': biases[0], 'dilation': dilations[0],
       'padding': paddings[0]}, [-2, 0]),
     ({'weight': weights[1], 'length': lengths[1],
       'bias': biases[1], 'dilation': dilations[1],
       'padding': paddings[1]}, [0, 0]),
     ({'weight': weights[2], 'length': lengths[2],
       'bias': biases[2], 'dilation': dilations[2],
       'padding': paddings[2]}, [2, 1])]
)
def test_apply_one_kernel_one_sample(params, arr_desired):
    """Test the convolutions applied to one time series."""
    arr_actual = apply_one_kernel_one_sample(
        x=X[0], n_timestamps=7, **params)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize(
    'params, arr_desired',
    [({'X': X, 'weights': np.zeros((3, 4)), 'lengths': np.array([2, 3, 4]),
       'biases': np.zeros(3), 'dilations': np.ones(3, dtype='int64'),
       'paddings': np.zeros(3, dtype='int64')}, np.zeros((2, 6))),

     ({'X': np.zeros((2, 7)), 'weights': np.ones((3, 4)),
       'lengths': np.array([2, 3, 4]), 'biases': np.zeros(3),
       'dilations': np.ones(3, dtype='int64'),
       'paddings': np.zeros(3, dtype='int64')}, np.zeros((2, 6))),

     ({'X': np.zeros((2, 7)), 'weights': np.ones((3, 4)),
       'lengths': np.array([2, 3, 4]), 'biases': np.ones(3),
       'dilations': np.ones(3, dtype='int64'),
       'paddings': np.zeros(3, dtype='int64')}, np.ones((2, 6))),

     ({'X': X, 'weights': weights, 'lengths': lengths,
       'biases': biases, 'dilations': dilations, 'paddings': paddings},
      [[-2, 0, 0, 0, 2, 1], [-2, 0, 0, 0, 5, 1]]),

     ({'X': X[:1], 'weights': weights, 'lengths': lengths,
       'biases': biases, 'dilations': dilations, 'paddings': paddings},
      [[-2, 0, 0, 0, 2, 1]]),

     ({'X': X[1:], 'weights': weights, 'lengths': lengths,
       'biases': biases, 'dilations': dilations, 'paddings': paddings},
      [[-2, 0, 0, 0, 5, 1]])]
)
def test_apply_all_kernels(params, arr_desired):
    """Test the application of all kernels to all samples."""
    arr_actual = apply_all_kernels(**params)
    np.testing.assert_allclose(arr_actual, arr_desired, atol=1e-5, rtol=0)


@pytest.mark.parametrize(
    'params, error, err_msg',
    [({'n_kernels': 'yolo'}, TypeError,
      "'n_kernels' must be an integer (got yolo)."),

     ({'kernel_sizes': {0: 1}}, TypeError,
      "'kernel_sizes' must be a list, a tuple or an array (got {0: 1})."),

     ({'kernel_sizes': np.arange(1)}, ValueError,
      "All the values in 'kernel_sizes' must be greater than or equal to 1 "
      "(0 < 1)."),

     ({'kernel_sizes': np.arange(3)}, ValueError,
      "All the values in 'kernel_sizes' must be greater than or equal to 1 "
      "(0 < 1)."),

     ({'kernel_sizes': np.arange(1, 9)}, ValueError,
      "All the values in 'kernel_sizes' must be lower than or equal to "
      "'n_timestamps' (8 > 7)."),

     ({'kernel_sizes': np.arange(1, 14)}, ValueError,
      "All the values in 'kernel_sizes' must be lower than or equal to "
      "'n_timestamps' (13 > 7).")]
)
def test_parameter_check(params, error, err_msg):
    """Test parameter validation."""
    clf = ROCKET(**params)
    with pytest.raises(error, match=re.escape(err_msg)):
        clf.fit(X)



"""Code for utility tools."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from numpy.lib.stride_tricks import as_strided
from numba import njit
from sklearn.utils import check_array


def segmentation(ts_size, window_size, overlapping=False, n_segments=None):
    """Compute the indices for Piecewise Agrgegate Approximation.

    Parameters
    ----------
    ts_size : int
        The size of the time series.

    window_size : int
        The size of the window.

    overlapping : bool (default = False)
        If True, overlapping windows may be used. If False, non-overlapping
        are used.

    n_segments : int or None (default = None)
        The number of windows. If None, the number is automatically
        computed using ``window_size``.

    Returns
    -------
    start : array
        The lower bound for each window.

    end : array
        The upper bound for each window.

    size : int
        The size of ``start``.

    Examples
    --------
    >>> from pyts.utils import segmentation
    >>> start, end, size = segmentation(ts_size=12, window_size=3)
    >>> print(start)
    [0 3 6 9]
    >>> print(end)
    [ 3  6  9 12]
    >>> size
    4

    """
    if not isinstance(ts_size, (int, np.integer)):
        raise TypeError("'ts_size' must be an integer.")
    if not ts_size >= 2:
        raise ValueError("'ts_size' must be an integer greater than or equal "
                         "to 2 (got {0}).".format(ts_size))
    if not isinstance(window_size, (int, np.integer)):
        raise TypeError("'window_size' must be an integer.")
    if not window_size >= 1:
        raise ValueError("'window_size' must be an integer greater than or "
                         "equal to 1 (got {0}).".format(window_size))
    if not window_size <= ts_size:
        raise ValueError("'window_size' must be lower than or equal to "
                         "'ts_size' ({0} > {1}).".format(window_size, ts_size))
    if not (n_segments is None or isinstance(n_segments, (int, np.integer))):
        raise TypeError("'n_segments' must be None or an integer.")
    if isinstance(n_segments, (int, np.integer)):
        if not n_segments >= 2:
            raise ValueError(
                "If 'n_segments' is an integer, it must be greater than or "
                "equal to 2 (got {0}).".format(n_segments)
            )
        if not n_segments <= ts_size:
            raise ValueError(
                "If 'n_segments' is an integer, it must be lower than or "
                "equal to 'ts_size' ({0} > {1}).".format(n_segments, ts_size)
            )

    if n_segments is None:
        quotient, remainder = divmod(ts_size, window_size)
        n_segments = quotient if remainder == 0 else quotient + 1

    if not overlapping:
        bounds = np.linspace(0, ts_size, n_segments + 1).astype('int64')
        start = bounds[:-1]
        end = bounds[1:]
        size = start.size
        return start, end, size
    else:
        n_overlapping = (n_segments * window_size) - ts_size
        n_overlaps = n_segments - 1
        overlaps = np.linspace(0, n_overlapping,
                               n_overlaps + 1).astype('int64')
        bounds = np.arange(0, (n_segments + 1) * window_size, window_size)
        start = bounds[:-1] - overlaps
        end = bounds[1:] - overlaps
        size = start.size
        return start, end, size


@njit()
def _windowed_view(X, n_samples, n_timestamps, window_size, window_step):
    overlap = window_size - window_step
    shape_new = (n_samples,
                 (n_timestamps - overlap) // window_step,
                 window_size // 1)
    s0, s1 = X.strides
    strides_new = (s0, window_step * s1, s1)
    return as_strided(X, shape=shape_new, strides=strides_new)


def windowed_view(X, window_size, window_step=1):
    """Return a windowed view of a 2D array.

    Parameters
    ----------
    X : array-like, shape = (n_samples, n_timestamps)
        Input data.

    window_size : int
        The size of the window. It must be between 1 and ``n_timestamps``.

    window_step : int (default = 1)
        The step of the sliding window

    Returns
    -------
    X_new : array, shape = (n_samples, n_windows, window_size)
        Windowed view of the input data. ``n_windows`` is computed as
        ``(n_timestamps - window_size + window_step) // window_step``.

    Examples
    --------
    >>> import numpy as np
    >>> from pyts.utils import windowed_view
    >>> windowed_view(np.arange(6).reshape(1, -1), window_size=2)
    array([[[0, 1],
            [1, 2],
            [2, 3],
            [3, 4],
            [4, 5]]])

    """
    X = check_array(X, dtype=None)
    n_samples, n_timestamps = X.shape

    if not isinstance(window_size, (int, np.integer)):
        raise TypeError("'window_size' must be an integer.")
    if not 1 <= window_size <= n_timestamps:
        raise ValueError("'window_size' must be an integer between 1 and "
                         "n_timestamps.")
    if not isinstance(window_step, (int, np.integer)):
        raise TypeError("'window_step' must be an integer.")
    if not 1 <= window_step <= n_timestamps:
        raise ValueError("'window_step' must be an integer between 1 and "
                         "n_timestamps.")

    return _windowed_view(X, n_samples, n_timestamps, window_size, window_step)



"""Functions to load datasets."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import os
from .uea import _load_uea_dataset
from .ucr import _load_ucr_dataset


def _load_dataset(name, archive, return_X_y):
    r"""Load and return dataset.

    Parameters
    ----------
    name : str
        Name of the dataset.

    archive : 'UCR' or 'UEA'
        Archive the dataset belongs to.

    return_X_y : bool
        If True, return
        ``(data_train, data_test, target_train, target_test)`` instead of a
        Bunch object.

    Returns
    -------
    data : Bunch
        Dictionary-like object, with attributes:

        data_train : array of floats
            The time series in the training set.
        data_test : array of floats
            The time series in the test set.
        target_train : array of integers
            The classification labels in the training set.
        target_test : array of integers
            The classification labels in the test set.
        DESCR : str
            The full description of the dataset.
        url : str
            The url of the dataset.

    (data_train, data_test, target_train, target_test) : tuple if ``return_X_y`` is True

    """  # noqa: E501
    module_path = os.path.dirname(__file__)
    folder = os.path.join(module_path, 'cached_datasets', archive, '')
    if archive == 'UCR':
        bunch = _load_ucr_dataset(name, folder)
    else:
        bunch = _load_uea_dataset(name, folder)
    if return_X_y:
        return (bunch.data_train, bunch.data_test,
                bunch.target_train, bunch.target_test)
    return bunch


def load_basic_motions(return_X_y=False):
    r"""Load and return the Basic Motions dataset.

    The data was generated as part of a student project where four students
    performed four activities whilst wearing a smart watch. The watch collects
    3D accelerometer and a 3D gyroscope It consists of four classes, which are
    walking, resting, running and badminton. Participants were required to
    record motion a total of five times, and the data is sampled once every
    tenth of a second, for a ten second period.

    ================   ==============
    Training samples               40
    Test samples                   40
    Dimensionality                  6
    Timestamps                    100
    Classes                         4
    ================   ==============

    Parameters
    ----------
    return_X_y : bool (default = False)
        If True, return
        ``(data_train, data_test, target_train, target_test)`` instead of a
        Bunch object.

    Returns
    -------
    data : Bunch
        Dictionary-like object, with attributes:

        data_train : array of floats
            The time series in the training set.
        data_test : array of floats
            The time series in the test set.
        target_train : array of integers
            The classification labels in the training set.
        target_test : array of integers
            The classification labels in the test set.
        DESCR : str
            The full description of the dataset.
        url : str
            The url of the dataset.

    (data_train, data_test, target_train, target_test) : tuple if ``return_X_y`` is True

    References
    ----------
    .. [1] `UCR archive entry for the PigCVP dataset
           <http://www.timeseriesclassification.com/description.php?Dataset=BasicMotions>`_

    Examples
    --------
    >>> from pyts.datasets import load_basic_motions
    >>> bunch = load_basic_motions()
    >>> bunch.data_train.shape
    (40, 6, 100)
    >>> X_train, X_test, y_train, y_test = load_basic_motions(return_X_y=True)
    >>> X_train.shape
    (40, 6, 100)

    """  # noqa: E501
    return _load_dataset('BasicMotions', 'UEA', return_X_y)


def load_coffee(return_X_y=False):
    r"""Load and return the Coffee dataset.

    Food spectrographs are used in chemometrics to classify food types, a task
    that has obvious applications in food safety and quality assurance. The
    coffee data set is a two class problem to distinguish between Robusta and
    Aribica coffee beans.

    ================   ==============
    Training samples               28
    Test samples                   28
    Timestamps                    286
    Classes                         2
    ================   ==============

    Parameters
    ----------
    return_X_y : bool (default = False)
        If True, return
        ``(data_train, data_test, target_train, target_test)`` instead of a
        Bunch object.

    Returns
    -------
    data : Bunch
        Dictionary-like object, with attributes:

        data_train : array of floats
            The time series in the training set.
        data_test : array of floats
            The time series in the test set.
        target_train : array of integers
            The classification labels in the training set.
        target_test : array of integers
            The classification labels in the test set.
        DESCR : str
            The full description of the dataset.
        url : str
            The url of the dataset.

    (data_train, data_test, target_train, target_test) : tuple if ``return_X_y`` is True

    References
    ----------
    .. [1] R. Briandet, E.K. Kemsley, and R.H. Wilson, "Discrimination of
           Arabica and Robusta in Instant Coffee by Fourier Transform Infrared
           Spectroscopy and Chemometrics". Journal of Agricultural and Food
           Chemistry (1996).

    .. [2] A. Bagnall, L. Davis, J. Hills and J. Lines, "Transformation Based
           Ensembles for Time Series Classification". SDM (2012).

    .. [3] `UCR archive entry for the PigCVP dataset
           <http://www.timeseriesclassification.com/description.php?Dataset=Coffee>`_

    Examples
    --------
    >>> from pyts.datasets import load_coffee
    >>> bunch = load_coffee()
    >>> bunch.data_train.shape
    (28, 286)
    >>> X_train, X_test, y_train, y_test = load_coffee(return_X_y=True)
    >>> X_train.shape
    (28, 286)

    """  # noqa: E501
    return _load_dataset('Coffee', 'UCR', return_X_y)


def load_gunpoint(return_X_y=False):
    r"""Load and return the GunPoint dataset.

    This dataset involves one female actor and one male actor making a motion
    with their hand. The two classes are: Gun-Draw and Point: For Gun-Draw the
    actors have their hands by their sides. They draw a replicate gun from a
    hip-mounted holster, point it at a target for approximately one second,
    then return the gun to the holster, and their hands to their sides. For
    Point the actors have their gun by their sides. They point with their index
    fingers to a target for approximately one second, and then return their
    hands to their sides. For both classes, we tracked the centroid of the
    actor's right hands in both X- and Y-axes, which appear to be highly
    correlated. The data in the archive is just the X-axis.

    ================   ==============
    Training samples               50
    Test samples                  150
    Timestamps                    150
    Classes                         2
    ================   ==============

    Parameters
    ----------
    return_X_y : bool (default = False)
        If True, return
        ``(data_train, data_test, target_train, target_test)`` instead of a
        Bunch object.

    Returns
    -------
    data : Bunch
        Dictionary-like object, with attributes:

        data_train : array of floats
            The time series in the training set.
        data_test : array of floats
            The time series in the test set.
        target_train : array of integers
            The classification labels in the training set.
        target_test : array of integers
            The classification labels in the test set.
        DESCR : str
            The full description of the dataset.
        url : str
            The url of the dataset.

    (data_train, data_test, target_train, target_test) : tuple if ``return_X_y`` is True

    References
    ----------
    .. [1] `UCR archive entry for the PigCVP dataset
           <http://www.timeseriesclassification.com/description.php?Dataset=GunPoint>`_

    Examples
    --------
    >>> from pyts.datasets import load_gunpoint
    >>> bunch = load_gunpoint()
    >>> bunch.data_train.shape
    (50, 150)
    >>> X_train, X_test, y_train, y_test = load_gunpoint(return_X_y=True)
    >>> X_train.shape
    (50, 150)

    """  # noqa: E501
    return _load_dataset('GunPoint', 'UCR', return_X_y)


def load_pig_central_venous_pressure(return_X_y=False):
    r"""Load and return the PigCVP dataset.

    In the test set, a class is represented by four examples, the second and
    third 2000 data points of the before time series and the second and third
    2000 data points of the after time series. Data created by Mathieu
    Guillame-Bert et al. Data edited by Shaghayegh Gharghabi and Eamonn Keogh.

    ================   ==============
    Training samples              104
    Test samples                  208
    Timestamps                   2000
    Classes                        52
    ================   ==============

    Parameters
    ----------
    return_X_y : bool (default = False)
        If True, return
        ``(data_train, data_test, target_train, target_test)`` instead of a
        Bunch object.

    Returns
    -------
    data : Bunch
        Dictionary-like object, with attributes:

        data_train : array of floats
            The time series in the training set.
        data_test : array of floats
            The time series in the test set.
        target_train : array of integers
            The classification labels in the training set.
        target_test : array of integers
            The classification labels in the test set.
        DESCR : str
            The full description of the dataset.
        url : str
            The url of the dataset.

    (data_train, data_test, target_train, target_test) : tuple if ``return_X_y`` is True

    References
    ----------
    .. [1] M. Guillame-Bert and A. Dubrawski, "Classification of Time Sequences
           using Graphs of Temporal Constraints". Journal of Machine Learning
           Research, 2017.

    .. [2] `UCR archive entry for the PigCVP dataset
           <http://www.timeseriesclassification.com/description.php?Dataset=PigCVP>`_

    Examples
    --------
    >>> from pyts.datasets import load_pig_central_venous_pressure
    >>> bunch = load_pig_central_venous_pressure()
    >>> bunch.data_train.shape
    (104, 2000)
    >>> X_train, X_test, y_train, y_test = load_pig_central_venous_pressure(
    ...    return_X_y=True)
    >>> X_train.shape
    (104, 2000)

    """  # noqa: E501
    return _load_dataset('PigCVP', 'UCR', return_X_y)



"""Testing for the UCR utility functions."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import pytest
import re
from pyts.datasets import fetch_ucr_dataset, ucr_dataset_info, ucr_dataset_list
from pyts.datasets.ucr import (_correct_ucr_name_description,
                               _correct_ucr_name_download)


@pytest.mark.parametrize(
    'dataset, err_msg',
    [('Hey',
      "Hey is not a valid name. The list of available names can be obtained "
      "by calling the 'pyts.datasets.ucr_dataset_list' function."),

     (['Hey', 'ACSF1'],
      "The following names are not valid: ['Hey']. The list of available "
      "names can be obtained by calling the "
      "'pyts.datasets.ucr_dataset_list' function."),

     (['Hey', 'Hi', 'ACSF1'],
      "The following names are not valid: ['Hey' 'Hi']. The list of "
      "available names can be obtained by calling the "
      "'pyts.datasets.ucr_dataset_list' function.")]
)
def test_parameter_check_uea_dataset_info(dataset, err_msg):
    """Test parameter validation."""
    with pytest.raises(ValueError, match=re.escape(err_msg)):
        ucr_dataset_info(dataset)


@pytest.mark.parametrize(
    'dataset, length_expected',
    [(None, 128),
     ('CBF', 5),
     (['CBF'], 1),
     (['CBF', 'Beef'], 2)]
)
def test_dictionary_length_uea_dataset_info(dataset, length_expected):
    """Test that the length of the dictionart is the expected one."""
    assert len(ucr_dataset_info(dataset)) == length_expected


@pytest.mark.parametrize(
    'dataset',
    [None,
     'Computers',
     ['Computers'],
     ['Computers', 'Chinatown']]
)
def test_dictionary_keys_ucr_dataset_info(dataset):
    """Test that the length of the dictionart is the expected one."""
    keys_expected = ['n_classes', 'n_timestamps', 'test_size', 'train_size',
                     'type']
    dictionary = ucr_dataset_info(dataset)
    if 'train_size' in dictionary.keys():
        assert sorted(list(dictionary.keys())) == keys_expected
    else:
        for key in dictionary.keys():
            assert sorted(list(dictionary[key].keys())) == keys_expected


def test_length_ucr_dataset_list():
    """Test that the length of the list is equal to the number of datasets."""
    assert len(ucr_dataset_list()) == 128


@pytest.mark.parametrize(
    'dataset, output',
    [('CinCECGtorso', 'CinCECGTorso'),
     ('MixedShapes', 'MixedShapesRegularTrain'),
     ('NonInvasiveFetalECGThorax1', 'NonInvasiveFatalECGThorax1'),
     ('NonInvasiveFetalECGThorax2', 'NonInvasiveFatalECGThorax2'),
     ('StarlightCurves', 'StarLightCurves'),
     ('Hey', 'Hey'),
     ('Hello World', 'Hello World')]
)
def test_correct_ucr_name_download(dataset, output):
    """Test that the results are the expected ones."""
    assert _correct_ucr_name_download(dataset) == output


@pytest.mark.parametrize(
    'dataset, output',
    [('CinCECGTorso', 'CinCECGtorso'),
     ('MixedShapesRegularTrain', 'MixedShapes'),
     ('NonInvasiveFatalECGThorax1', 'NonInvasiveFetalECGThorax1'),
     ('NonInvasiveFatalECGThorax2', 'NonInvasiveFetalECGThorax2'),
     ('StarLightCurves', 'StarlightCurves'),
     ('Hey', 'Hey'),
     ('Hello World', 'Hello World')]
)
def test_correct_ucr_name_description(dataset, output):
    """Test that the results are the expected ones."""
    assert _correct_ucr_name_description(dataset) == output


def test_fetch_cached_ucr_dataset():
    """Test that a cached dataset can be loaded using 'fetch_ucr_dataset'."""
    res = fetch_ucr_dataset('GunPoint', use_cache=True)
    assert res.data_train.shape == (50, 150)
    assert res.data_test.shape == (150, 150)
    assert res.target_train.shape == (50,)
    assert res.target_test.shape == (150,)



"""Utility class for multivariate time series transformation."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from scipy.sparse import csr_matrix, hstack
from sklearn.base import BaseEstimator, clone
from sklearn.utils.validation import check_is_fitted
from ...base import MultivariateTransformerMixin
from ..utils import check_3d_array


class MultivariateTransformer(BaseEstimator, MultivariateTransformerMixin):
    r"""Transformer for multivariate time series.

    It provides a convenient class to transform multivariate time series with
    transformers that can only deal with univariate time series.

    Parameters
    ----------
    estimator : estimator object or list thereof
        Transformer. If one estimator is provided, it is cloned and each clone
        transforms one feature. If a list of estimators is provided, each
        estimator transforms one feature.

    flatten : bool (default = True)
        Affect shape of transform output. If True, ``transform``
        returns an array with shape (n_samples, \*). If False, the output of
        ``transform`` from each estimator must have the same shape and
        ``transform`` returns an array with shape (n_samples, n_features, \*).
        Ignored if the transformers return sparse matrices.

    Attributes
    ----------
    estimators_ : list of estimator objects
        The collection of fitted transformers.

    Examples
    --------
    >>> from pyts.datasets import load_basic_motions
    >>> from pyts.multivariate.transformation import MultivariateTransformer
    >>> from pyts.image import GramianAngularField
    >>> X, _, _, _ = load_basic_motions(return_X_y=True)
    >>> transformer = MultivariateTransformer(GramianAngularField(),
    ...                                       flatten=False)
    >>> X_new = transformer.fit_transform(X)
    >>> X_new.shape
    (40, 6, 100, 100)

    """

    def __init__(self, estimator, flatten=True):
        self.estimator = estimator
        self.flatten = flatten

    def fit(self, X, y=None):
        """Pass.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features, n_timestamps)
            Multivariate time series.

        y : None or array-like, shape = (n_samples,) (default = None)
            Class labels.

        Returns
        -------
        self : object

        """
        X = check_3d_array(X)
        _, n_features, _ = X.shape
        self._check_params(n_features)
        for i, transformer in enumerate(self.estimators_):
            transformer.fit(X[:, i, :], y)
        return self

    def transform(self, X):
        r"""Apply transform to each feature.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features, n_timestamps)
            Multivariate time series.

        Returns
        -------
        X_new : array, shape = (n_samples, *) or (n_samples, n_features, *)
            Transformed time series.

        """
        X = check_3d_array(X)
        n_samples, _, _ = X.shape
        check_is_fitted(self, 'estimators_')

        X_transformed = [transformer.transform(X[:, i, :])
                         for i, transformer in enumerate(self.estimators_)]
        all_sparse = np.all([isinstance(X_transformed_i, csr_matrix)
                             for X_transformed_i in X_transformed])
        if all_sparse:
            X_new = hstack(X_transformed)
        else:
            X_new = [self._convert_to_array(X_transformed_i)
                     for X_transformed_i in X_transformed]
            ndims = [X_new_i.ndim for X_new_i in X_new]
            shapes = [X_new_i.shape for X_new_i in X_new]
            one_dim = (np.unique(ndims).size == 1)
            if one_dim:
                one_shape = np.unique(shapes, axis=0).shape[0] == 1
            else:
                one_shape = False
            if (not one_shape) or self.flatten:
                X_new = [X_new_i.reshape(n_samples, -1) for X_new_i in X_new]
                X_new = np.concatenate(X_new, axis=1)
            else:
                X_new = np.asarray(X_new)
                axes = [1, 0] + [i for i in range(2, X_new.ndim)]
                X_new = np.transpose(X_new, axes=axes)
        return X_new

    def _check_params(self, n_features):
        """Check parameters."""
        transformer = (isinstance(self.estimator, BaseEstimator)
                       and hasattr(self.estimator, 'transform'))
        if transformer:
            self.estimators_ = [clone(self.estimator)
                                for _ in range(n_features)]

        elif isinstance(self.estimator, list):
            if len(self.estimator) != n_features:
                raise ValueError(
                    "If 'estimator' is a list, its length must be equal to "
                    "the number of features ({0} != {1})"
                    .format(len(self.estimator), n_features)
                )
            for i, estimator in enumerate(self.estimator):
                if not (isinstance(estimator, BaseEstimator)
                        and hasattr(estimator, 'transform')):
                    raise ValueError("Estimator {} must be a transformer."
                                     .format(i))
            self.estimators_ = self.estimator

        else:
            raise TypeError(
                "'estimator' must be a transformer that inherits from "
                "sklearn.base.BaseEstimator or a list thereof.")

    @staticmethod
    def _convert_to_array(X):
        """Convert the input data to an array if necessary."""
        if isinstance(X, csr_matrix):
            return X.A
        elif isinstance(X, np.ndarray):
            return X
        else:
            raise ValueError('Unexpected type for X: {}.'
                             .format(type(X).__name__))



"""
Utility functions for the UCR multivariate time series classification
archive.
"""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
import os
import pickle
from scipy.io.arff import loadarff
from sklearn.utils import Bunch
from urllib.request import urlretrieve
import zipfile


def _correct_ucr_name_download(dataset):
    if dataset == 'CinCECGtorso':
        return 'CinCECGTorso'
    elif dataset == 'MixedShapes':
        return 'MixedShapesRegularTrain'
    elif dataset == 'NonInvasiveFetalECGThorax1':
        return 'NonInvasiveFatalECGThorax1'
    elif dataset == 'NonInvasiveFetalECGThorax2':
        return 'NonInvasiveFatalECGThorax2'
    elif dataset == 'StarlightCurves':
        return 'StarLightCurves'
    else:
        return dataset


def _correct_ucr_name_description(dataset):
    if dataset == 'CinCECGTorso':
        return 'CinCECGtorso'
    elif dataset == 'MixedShapesRegularTrain':
        return 'MixedShapes'
    elif dataset == 'NonInvasiveFatalECGThorax1':
        return 'NonInvasiveFetalECGThorax1'
    elif dataset == 'NonInvasiveFatalECGThorax2':
        return 'NonInvasiveFetalECGThorax2'
    elif dataset == 'StarLightCurves':
        return 'StarlightCurves'
    else:
        return dataset


def ucr_dataset_list():
    """List of available UCR datasets.

    Returns
    -------
    datasets : list
        List of available datasets from the UCR Time Series
        Classification Archive.

    References
    ----------
    .. [1] `List of datasets on the UEA & UCR archive
           <http://www.timeseriesclassification.com/dataset.php>`_

    Examples
    --------
    >>> from pyts.datasets import ucr_dataset_list
    >>> ucr_dataset_list()[:3]
    ['ACSF1', 'Adiac', 'AllGestureWiimoteX']

    """
    module_path = os.path.dirname(__file__)
    finfo = os.path.join(module_path, 'info', 'ucr.pickle')
    dictionary = pickle.load(open(finfo, 'rb'))
    datasets = sorted(dictionary.keys())
    return datasets


def ucr_dataset_info(dataset=None):
    """Information about the UCR datasets.

    Parameters
    ----------
    dataset : str, list of str or None (default = None)
        The data sets for which the information will be returned.
        If None, the information for all the datasets is returned.

    Returns
    -------
    dictionary : dict
        Dictionary with the information for each dataset.

    References
    ----------
    .. [1] `List of datasets on the UEA & UCR archive
           <http://www.timeseriesclassification.com/dataset.php>`_

    Examples
    --------
    >>> from pyts.datasets import ucr_dataset_info
    >>> ucr_dataset_info('Adiac')['n_classes']
    37

    """
    module_path = os.path.dirname(__file__)
    finfo = os.path.join(module_path, 'info', 'ucr.pickle')
    dictionary = pickle.load(open(finfo, 'rb'))
    datasets = list(dictionary.keys())

    if dataset is None:
        return dictionary
    elif isinstance(dataset, str):
        if dataset not in datasets:
            raise ValueError(
                "{0} is not a valid name. The list of available names "
                "can be obtained by calling the "
                "'pyts.datasets.ucr_dataset_list' function."
                .format(dataset)
            )
        else:
            return dictionary[dataset]
    elif isinstance(dataset, (list, tuple, np.ndarray)):
        dataset = np.asarray(dataset)
        invalid_datasets = np.setdiff1d(dataset, datasets)
        if invalid_datasets.size > 0:
            raise ValueError(
                "The following names are not valid: {0}. The list of "
                "available names can be obtained by calling the "
                "'pyts.datasets.ucr_dataset_list' function."
                .format(invalid_datasets)
            )
        else:
            info = {}
            for data in dataset:
                info[data] = dictionary[data]
            return info


def fetch_ucr_dataset(dataset, use_cache=True, data_home=None,
                      return_X_y=False):
    r"""Fetch dataset from UCR TSC Archive by name.

    Fetched data sets are automatically saved in the
    ``pyts/datasets/_cached_datasets`` folder. To avoid
    downloading the same data set several times, it is
    highly recommended not to change the default values
    of ``use_cache`` and ``path``.

    Parameters
    ----------
    dataset : str
        Name of the dataset.

    use_cache : bool (default = True)
        If True, look if the data set has already been fetched
        and load the fetched version if it is the case. If False,
        download the data set from the UCR Time Series Classification
        Archive.

    data_home : None or str (default = None)
        The path of the folder containing the cached data set.
        If None, the ``pyts.datasets.cached_datasets/UCR/`` folder is
        used. If the data set is not found, it is downloaded and cached
        in this path.

    return_X_y : bool (default = False)
        If True, returns ``(data_train, data_test, target_train, target_test)``
        instead of a Bunch object. See below for more information about the
        `data` and `target` object.

    Returns
    -------
    data : Bunch
        Dictionary-like object, with attributes:

        data_train : array of floats
            The time series in the training set.
        data_test : array of floats
            The time series in the test set.
        target_train : array of integers
            The classification labels in the training set.
        target_test : array of integers
            The classification labels in the test set.
        DESCR : str
            The full description of the dataset.
        url : str
            The url of the dataset.

    (data_train, data_test, target_train, target_test) : tuple if ``return_X_y`` is True

    Notes
    -----
    Missing values are represented as NaN's.

    References
    ----------
    .. [1] H. A. Dau et al, "The UCR Time Series Archive".
           arXiv:1810.07758 [cs, stat], 2018.

    .. [2] A. Bagnall et al, "The UEA & UCR Time Series Classification
           Repository", www.timeseriesclassification.com.

    """  # noqa: E501
    if dataset not in ucr_dataset_list():
        raise ValueError(
            "{0} is not a valid name. The list of available names "
            "can be obtained with ``pyts.datasets.ucr_dataset_list()``"
            .format(dataset)
        )
    if data_home is None:
        import pyts
        home = '/'.join(pyts.__file__.split('/')[:-2]) + '/'
        relative_path = 'pyts/datasets/cached_datasets/UCR/'
        path = home + relative_path
    else:
        path = data_home
    if not os.path.exists(path):
        os.makedirs(path)

    correct_dataset = _correct_ucr_name_download(dataset)
    if use_cache and os.path.exists(path + correct_dataset):
        bunch = _load_ucr_dataset(correct_dataset, path=path)
    else:
        url = ("http://www.timeseriesclassification.com/Downloads/{0}.zip"
               .format(correct_dataset))
        filename = 'temp_{}'.format(correct_dataset)
        _ = urlretrieve(url, path + filename)
        zipfile.ZipFile(path + filename).extractall(path + correct_dataset)
        os.remove(path + filename)
        bunch = _load_ucr_dataset(correct_dataset, path)

    if return_X_y:
        return (bunch.data_train, bunch.data_test,
                bunch.target_train, bunch.target_test)
    return bunch


def _load_ucr_dataset(dataset, path):
    """Load a UCR data set from a local folder.

    Parameters
    ----------
    dataset : str
        Name of the dataset.

    path : str
        The path of the folder containing the cached data set.

    Returns
    -------
    data : Bunch
        Dictionary-like object, with attributes:

        data_train : array of floats
            The time series in the training set.
        data_test : array of floats
            The time series in the test set.
        target_train : array
            The classification labels in the training set.
        target_test : array
            The classification labels in the test set.
        DESCR : str
            The full description of the dataset.
        url : str
            The url of the dataset.

    Notes
    -----
    Padded values are represented as NaN's.

    """
    new_path = path + dataset + '/'
    try:
        with(open(new_path + dataset + '.txt', encoding='utf-8')) as f:
            description = f.read()
    except UnicodeDecodeError:
        with(open(new_path + dataset + '.txt', encoding='ISO-8859-1')) as f:
            description = f.read()
    try:
        data_train = np.genfromtxt(new_path + dataset + '_TRAIN.txt')
        data_test = np.genfromtxt(new_path + dataset + '_TEST.txt')

        X_train, y_train = data_train[:, 1:], data_train[:, 0]
        X_test, y_test = data_test[:, 1:], data_test[:, 0]

    except IndexError:
        train = loadarff(new_path + dataset + '_TRAIN.arff')
        test = loadarff(new_path + dataset + '_TEST.arff')

        data_train = np.asarray([train[0][name] for name in train[1].names()])
        X_train = data_train[:-1].T.astype('float64')
        y_train = data_train[-1]

        data_test = np.asarray([test[0][name] for name in test[1].names()])
        X_test = data_test[:-1].T.astype('float64')
        y_test = data_test[-1]

    try:
        y_train = y_train.astype('float64').astype('int64')
        y_test = y_test.astype('float64').astype('int64')
    except ValueError:
        y_train = y_train.astype(str)
        y_test = y_test.astype(str)

    bunch = Bunch(
        data_train=X_train, target_train=y_train,
        data_test=X_test, target_test=y_test,
        DESCR=description,
        url=("http://www.timeseriesclassification.com/"
             "description.php?Dataset={}".format(dataset))
    )

    return bunch



"""Code for Reccurence Plot."""

# Author: Johann Faouzi <johann.faouzi@gmail.com>
# License: BSD-3-Clause

import numpy as np
from math import ceil
from numpy.lib.stride_tricks import as_strided
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array
from ..base import UnivariateTransformerMixin


def _trajectories(X, dimension, time_delay):
    n_samples, n_timestamps = X.shape
    shape_new = (n_samples,
                 n_timestamps - (dimension - 1) * time_delay,
                 dimension)
    s0, s1 = X.strides
    strides_new = (s0, s1, time_delay * s1)
    return as_strided(X, shape=shape_new, strides=strides_new)


class RecurrencePlot(BaseEstimator, UnivariateTransformerMixin):  # noqa: D207
    r"""Recurrence Plot.

    A recurrence plot is an image representing the distances between
    trajectories extracted from the original time series.

    Parameters
    ----------
    dimension : int or float (default = 1)
        Dimension of the trajectory. If float, If float, it represents
        a percentage of the size of each time series and must be between
        0 and 1.

    time_delay : int or float (default = 1)
        Time gap between two back-to-back points of the trajectory. If
        float, If float, it represents a percentage of the size of each
        time series and must be between 0 and 1.

    threshold : float, 'point', 'distance' or None (default = None)
        Threshold for the minimum distance. If None, the recurrence plots
        are not binarized. If 'point', the threshold is computed such as
        `percentage` percents of the points are smaller than the threshold.
        If 'distance', the threshold is computed as the `percentage` of the
        maximum distance.

    percentage : int or float (default = 10)
        Percentage of black points if ``threshold='point'`` or percentage of
        maximum distance for threshold if ``threshold='distance'``.
        Ignored if ``threshold`` is a float or None.

    flatten : bool (default = False)
        If True, images are flattened to be one-dimensional.

    Notes
    -----
    Given a time series :math:`(x_1, \ldots, x_n)`, the extracted
    trajectories are

    .. math::

        \vec{x}_i = (x_i, x_{i + \tau}, \ldots, x_{i + (m - 1)\tau}), \quad
        \forall i \in \{1, \ldots, n - (m - 1)\tau \}

    where :math:`m` is the ``dimension`` of the trajectories and :math:`\tau`
    is the ``time_delay``. The recurrence plot, denoted :math:`R`, is the
    pairwise distance between the trajectories

    .. math::

        R_{i, j} = \Theta(\varepsilon - \| \vec{x}_i - \vec{x}_j \|), \quad
        \forall i,j \in \{1, \ldots, n - (m - 1)\tau \}

    where :math:`\Theta` is the Heaviside function and :math:`\varepsilon`
    is the ``threshold``.

    References
    ----------
    .. [1] J.-P Eckmann, S. Oliffson Kamphorst and D Ruelle, "Recurrence
           Plots of Dynamical Systems". Europhysics Letters (1987).

    Examples
    --------
    >>> from pyts.datasets import load_gunpoint
    >>> from pyts.image import RecurrencePlot
    >>> X, _, _, _ = load_gunpoint(return_X_y=True)
    >>> transformer = RecurrencePlot()
    >>> X_new = transformer.transform(X)
    >>> X_new.shape
    (50, 150, 150)

    """

    def __init__(self, dimension=1, time_delay=1,
                 threshold=None, percentage=10, flatten=False):
        self.dimension = dimension
        self.time_delay = time_delay
        self.threshold = threshold
        self.percentage = percentage
        self.flatten = flatten

    def fit(self, X=None, y=None):
        """Pass.

        Parameters
        ----------
        X
            Ignored
        y
            Ignored

        Returns
        -------
        self : object

        """
        return self

    def transform(self, X):
        """Transform each time series into a recurrence plot.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_timestamps)

        Returns
        -------
        X_new : array, shape = (n_samples, image_size, image_size)
            Recurrence plots. ``image_size`` is the number of
            trajectories and is equal to
            ``n_timestamps - (dimension - 1) * time_delay``.
            If ``flatten=True``, the shape is
            `(n_samples, image_size * image_size)`.

        """
        X = check_array(X)
        n_samples, n_timestamps = X.shape
        dimension, time_delay = self._check_params(n_timestamps)

        if dimension == 1:
            X_dist = np.abs(X[:, :, None] - X[:, None, :])
        else:
            X_traj = _trajectories(X, dimension, time_delay)
            X_dist = np.sqrt(
                np.sum((X_traj[:, None, :, :] - X_traj[:, :, None, :]) ** 2,
                       axis=3)
            )
        if self.threshold is None:
            X_rp = X_dist
        elif self.threshold == 'point':
            image_size = n_timestamps - (dimension - 1) * time_delay
            percents = np.percentile(
                np.reshape(X_dist, (n_samples, image_size * image_size)),
                self.percentage, axis=1
            )
            X_rp = X_dist < percents[:, None, None]
        elif self.threshold == 'distance':
            percents = self.percentage / 100 * np.max(X_dist, axis=(1, 2))
            X_rp = X_dist < percents[:, None, None]
        else:
            X_rp = X_dist < self.threshold

        if self.flatten:
            return X_rp.reshape(n_samples, -1).astype('float64')
        return X_rp.astype('float64')

    def _check_params(self, n_timestamps):
        if not isinstance(self.dimension,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'dimension' must be an integer or a float.")
        if isinstance(self.dimension, (int, np.integer)):
            if not 1 <= self.dimension <= n_timestamps:
                raise ValueError(
                    "If 'dimension' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.dimension)
                )
            dimension = self.dimension
        else:
            if not 0 < self.dimension < 1.:
                raise ValueError(
                    "If 'dimension' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 "
                    "(got {0}).".format(self.dimension)
                )
            dimension = ceil(self.dimension * n_timestamps)

        if not isinstance(self.time_delay,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'time_delay' must be an integer or a float.")
        if isinstance(self.time_delay, (int, np.integer)):
            if not 1 <= self.time_delay <= n_timestamps:
                raise ValueError(
                    "If 'time_delay' is an integer, it must be greater "
                    "than or equal to 1 and lower than or equal to "
                    "n_timestamps (got {0}).".format(self.time_delay)
                )
            time_delay = self.time_delay
        else:
            if not 0 < self.time_delay < 1.:
                raise ValueError(
                    "If 'time_delay' is a float, it must be greater "
                    "than 0 and lower than or equal to 1 "
                    "(got {0}).".format(self.time_delay)
                )
            time_delay = ceil(self.time_delay * n_timestamps)

        if n_timestamps - (dimension - 1) * time_delay < 1:
            raise ValueError("The number of trajectories must be positive. "
                             "Consider trying with smaller values for "
                             "'dimension' and 'time_delay'.")

        if (self.threshold is not None
            and self.threshold not in ['point', 'distance']
            and not isinstance(self.threshold,
                               (int, np.integer, float, np.floating))):
            raise TypeError("'threshold' must be either None, 'point', "
                            "'distance', a float or an integer.")
        threshold_number = isinstance(self.threshold,
                                      (int, np.integer, float, np.floating))
        if threshold_number and (self.threshold <= 0):
            raise ValueError("If 'threshold' is a float or an integer, "
                             "it must be greater than or equal to 0.")

        if not isinstance(self.percentage,
                          (int, np.integer, float, np.floating)):
            raise TypeError("'percentage' must be a float or an integer.")
        if not 0 <= self.percentage <= 100:
            raise ValueError("'percentage' must be between 0 and 100.")

        return dimension, time_delay



"""
The :mod:`pyts.image` module includes algorithms that transform times series
into images.
"""

from .gaf import GramianAngularField
from .mtf import MarkovTransitionField
from .recurrence import RecurrencePlot

__all__ = ['GramianAngularField', 'MarkovTransitionField', 'RecurrencePlot']
